@@ -51,14 +51,20 @@
 #endif
 
 #ifdef WIN32
-#include <winsock.h>
+#  include <winsock.h>
+
+#  ifndef R_OK
+#    define R_OK 04
+#  endif
+   /* we also need io.h for access() prototype */
+#  include <io.h>
 #else
-#include <sys/param.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <netdb.h>
-#endif /*WIN32*/
+#  include <sys/param.h>
+#  include <sys/socket.h>
+#  include <netinet/in.h>
+#  include <arpa/inet.h>
+#  include <netdb.h>
+#endif /* WIN32 */
 #include <fcntl.h>
 #include <stdio.h>
 #include <sasl.h>
@@ -110,10 +116,12 @@ typedef struct context {
     unsigned needsize;
 } context_t;
 
-#define SASL_GSSAPI_STATE_AUTHNEG 1
-#define SASL_GSSAPI_STATE_SSFCAP  2
-#define SASL_GSSAPI_STATE_SSFREQ  3
-#define SASL_GSSAPI_STATE_AUTHENTICATED    4
+enum {
+    SASL_GSSAPI_STATE_AUTHNEG = 1,
+    SASL_GSSAPI_STATE_SSFCAP = 2,
+    SASL_GSSAPI_STATE_SSFREQ = 3,
+    SASL_GSSAPI_STATE_AUTHENTICATED = 4
+};
 
 static void
 sasl_gss_disperr(context_t *context, char **outp, 
@@ -870,7 +878,6 @@ static const sasl_server_plug_t plugins[] =
   }
 };
 
-
 int 
 sasl_server_plug_init(sasl_utils_t *utils __attribute__((unused)), 
 		      int maxversion,
@@ -878,15 +885,36 @@ sasl_server_plug_init(sasl_utils_t *utils __attribute__((unused)),
 		      const sasl_server_plug_t **pluglist,
 		      int *plugcount)
 {
-  if (maxversion<GSSAPI_VERSION)
-    return SASL_BADVERS;
+#ifdef HAVE_GSSKRB5_REGISTER_ACCEPTOR_IDENTITY
+    const char *keytab;
+    unsigned int rl;
+#endif
 
-  *pluglist=plugins;
+    if (maxversion < GSSAPI_VERSION) {
+	return SASL_BADVERS;
+    }
 
-  *plugcount=1;  
-  *out_version=GSSAPI_VERSION;
+#ifdef HAVE_GSSKRB5_REGISTER_ACCEPTOR_IDENTITY
+    /* unfortunately, we don't check for readability of keytab if it's
+       the standard one, since we don't know where it is */
 
-  return SASL_OK;
+    utils->getopt(utils->getopt_context, "GSSAPI", "keytab", &keytab, &rl);
+    if (keytab != NULL) {
+	if (access(keytab, R_OK) != 0) {
+	    utils->log(NULL, SASL_LOG_ERR, "GSSAPI", SASL_FAIL, errno,
+		       "can't access keytab file %s: %m", keytab);
+	    return SASL_FAIL;
+	}
+
+	gsskrb5_register_acceptor_identity(keytab);
+    }
+#endif
+
+    *pluglist = plugins;
+    *plugcount = 1;  
+    *out_version = GSSAPI_VERSION;
+
+    return SASL_OK;
 }
 
 static int 
@@ -1058,56 +1086,6 @@ get_userid(sasl_client_params_t *params,char **userid,sasl_interact_t **prompt_n
   return result;
 }
 
-#if 0
-static int  /* borrowed from plain.c */
-get_authid(sasl_client_params_t *params,
-	   char **authid,
-	   sasl_interact_t **prompt_need)
-{
-  
-  int result;
-  sasl_getsimple_t *getauth_cb;
-  void *getauth_context;
-  sasl_interact_t *prompt;
-  const char *id;
-
-  /* see if we were given the authname in the prompt */
-  prompt=find_prompt(prompt_need,SASL_CB_AUTHNAME);
-  if (prompt!=NULL)
-  {
-      /* copy it */
-      *authid=params->utils->malloc(prompt->len+1);
-      if ((*authid)==NULL) return SASL_NOMEM;
-
-      strncpy(*authid, prompt->result, prompt->len+1);      
-      return SASL_OK;
-  }
-
-  /* Try to get the callback... */
-  result = params->utils->getcallback(params->utils->conn,
-				      SASL_CB_AUTHNAME,
-				      &getauth_cb,
-				      &getauth_context);
-  if (result == SASL_OK && getauth_cb) {
-    id = NULL;
-    result = getauth_cb(getauth_context,
-			SASL_CB_AUTHNAME,
-			&id,
-			NULL);
-    if (result != SASL_OK)
-      return result;
-    if (! id)
-      return SASL_BADPARAM;
-    *authid = params->utils->malloc(strlen(id) + 1);
-    if (! *authid)
-      return SASL_NOMEM;
-    strcpy(*authid, id);
-  }
-
-  return result;
-}
-#endif
-
 static int 
 sasl_gss_client_step (void *conn_context,
 		      sasl_client_params_t *params,
@@ -1130,6 +1108,11 @@ sasl_gss_client_step (void *conn_context,
   input_token->value = NULL; 
   input_token->length = 0;
 
+  if (!clientout && text->state == SASL_GSSAPI_STATE_AUTHNEG) {
+      /* initial client send not allowed */
+      return SASL_CONTINUE;
+  }
+
   *clientout = NULL;
   *clientoutlen = 0;
 
@@ -1291,9 +1274,20 @@ sasl_gss_client_step (void *conn_context,
 	    return SASL_BADPARAM;
 	}
 
-	/* need bits of layer */
-	allowed = secprops.max_ssf - external;
-	need = secprops.min_ssf - external;
+	/* need bits of layer -- sasl_ssf_t is unsigned so be careful */
+	if (secprops.max_ssf >= external) {
+	    allowed = secprops.max_ssf - external;
+	} else {
+	    allowed = 0;
+	}
+	if (secprops.min_ssf >= external) {
+	    need = secprops.min_ssf - external;
+	} else {
+	    /* good to go */
+	    need = 0;
+	}
+
+	/* bit mask of server support */
 	serverhas = ((char *)output_token->value)[0];
 
 	/* if client didn't set use strongest layer available */
